"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.Reconciler = undefined;

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }(); /**
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      * Copyright (c) 2017-present, Facebook, Inc.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      * All rights reserved.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      * This source code is licensed under the BSD-style license found in the
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      * LICENSE file in the root directory of this source tree. An additional grant
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      * of patent rights can be found in the PATENTS file in the same directory.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      */

var _realm = require("../realm.js");

var _modules = require("../utils/modules.js");

var _index = require("../values/index.js");

var _types = require("../serializer/types.js");

var _utils = require("./utils");

var _index2 = require("../methods/index.js");

var _invariant = require("../invariant.js");

var _invariant2 = _interopRequireDefault(_invariant);

var _singletons = require("../singletons.js");

var _errors = require("../errors.js");

var _branching = require("./branching.js");

var _components = require("./components.js");

var _errors2 = require("./errors.js");

var _completions = require("../completions.js");

var _logger = require("../utils/logger.js");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var Reconciler = exports.Reconciler = function () {
  function Reconciler(realm, moduleTracer, statistics, reactSerializerState, simpleClassComponents, branchReactComponentTrees) {
    _classCallCheck(this, Reconciler);

    this.realm = realm;
    this.moduleTracer = moduleTracer;
    this.statistics = statistics;
    this.reactSerializerState = reactSerializerState;
    this.simpleClassComponents = simpleClassComponents;
    this.logger = moduleTracer.modules.logger;
    this.branchReactComponentTrees = branchReactComponentTrees;
  }

  _createClass(Reconciler, [{
    key: "render",
    value: function render(componentType, props, context, isRoot, evaluatedRootNode) {
      var _this = this;

      return this.realm.wrapInGlobalEnv(function () {
        return _this.realm.evaluatePure(function () {
          return (
            // TODO: (sebmarkbage): You could use the return value of this to detect if there are any mutations on objects other
            // than newly created ones. Then log those to the error logger. That'll help us track violations in
            // components. :)
            _this.realm.evaluateForEffects(function () {
              // initialProps and initialContext are created from Flow types from:
              // - if a functional component, the 1st and 2nd paramater of function
              // - if a class component, use this.props and this.context
              // if there are no Flow types for props or context, we will throw a
              // FatalError, unless it's a functional component that has no paramater
              // i.e let MyComponent = () => <div>Hello world</div>
              try {
                var initialProps = props || (0, _components.getInitialProps)(_this.realm, componentType);
                var initialContext = context || (0, _components.getInitialContext)(_this.realm, componentType);

                var _renderComponent2 = _this._renderComponent(componentType, initialProps, initialContext, "ROOT", null, evaluatedRootNode),
                    result = _renderComponent2.result;

                _this.statistics.optimizedTrees++;
                return result;
              } catch (error) {
                // if we get an error and we're not dealing with the root
                // rather than throw a FatalError, we log the error as a warning
                // and continue with the other tree roots
                // TODO: maybe control what levels gets treated as warning/error?
                if (!isRoot) {
                  _this.logger.logWarning(componentType, "__registerReactComponentRoot() React component tree (branch) failed due to - " + error.message);
                  return _this.realm.intrinsics.undefined;
                }
                // if there was a bail-out on the root component in this reconcilation process, then this
                // should be an invariant as the user has explicitly asked for this component to get folded
                if (error instanceof _completions.Completion) {
                  _this.logger.logCompletion(error);
                  throw error;
                } else if (error instanceof _errors2.ExpectedBailOut) {
                  var diagnostic = new _errors.CompilerDiagnostic("__registerReactComponentRoot() React component tree (root) failed due to - " + error.message, _this.realm.currentLocation, "PP0020", "FatalError");
                  _this.realm.handleError(diagnostic);
                  if (_this.realm.handleError(diagnostic) === "Fail") throw new _errors.FatalError();
                }
                throw error;
              }
            },
            /*state*/null, "react component: " + componentType.getName())
          );
        });
      });
    }
  }, {
    key: "_queueNewComponentTree",
    value: function _queueNewComponentTree(rootValue, evaluatedNode) {
      var nested = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : false;
      var props = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : null;
      var context = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : null;

      (0, _invariant2.default)(rootValue instanceof _index.ECMAScriptSourceFunctionValue || rootValue instanceof _index.AbstractValue);
      this.branchReactComponentTrees.push({
        context: context,
        evaluatedNode: evaluatedNode,
        nested: nested,
        props: props,
        rootValue: rootValue
      });
    }
  }, {
    key: "_renderComplexClassComponent",
    value: function _renderComplexClassComponent(componentType, props, context, classMetadata, branchStatus, branchState, evaluatedNode) {
      if (branchStatus !== "ROOT") {
        this._queueNewComponentTree(componentType, evaluatedNode);
        evaluatedNode.status = "NEW_TREE";
        throw new _errors2.NewComponentTreeBranch();
      }
      // create a new instance of this React class component
      var instance = (0, _components.createClassInstance)(this.realm, componentType, props, context, classMetadata);
      // get the "render" method off the instance
      var renderMethod = (0, _index2.Get)(this.realm, instance, "render");
      (0, _invariant2.default)(renderMethod instanceof _index.ECMAScriptSourceFunctionValue && renderMethod.$Call, "Expected render method to be a FunctionValue with $Call method");
      // the render method doesn't have any arguments, so we just assign the context of "this" to be the instance
      return renderMethod.$Call(instance, []);
    }
  }, {
    key: "_renderSimpleClassComponent",
    value: function _renderSimpleClassComponent(componentType, props, context, branchStatus, branchState) {
      // create a new simple instance of this React class component
      var instance = (0, _components.createSimpleClassInstance)(this.realm, componentType, props, context);
      // get the "render" method off the instance
      var renderMethod = (0, _index2.Get)(this.realm, instance, "render");
      (0, _invariant2.default)(renderMethod instanceof _index.ECMAScriptSourceFunctionValue && renderMethod.$Call, "Expected render method to be a FunctionValue with $Call method");
      // the render method doesn't have any arguments, so we just assign the context of "this" to be the instance
      return renderMethod.$Call(instance, []);
    }
  }, {
    key: "_renderFunctionalComponent",
    value: function _renderFunctionalComponent(componentType, props, context) {
      (0, _invariant2.default)(componentType.$Call, "Expected componentType to be a FunctionValue with $Call method");
      return componentType.$Call(this.realm.intrinsics.undefined, [props, context]);
    }
  }, {
    key: "_getClassComponentMetadata",
    value: function _getClassComponentMetadata(componentType, props, context) {
      if (this.realm.react.classComponentMetadata.has(componentType)) {
        var _classMetadata = this.realm.react.classComponentMetadata.get(componentType);
        (0, _invariant2.default)(_classMetadata);
        return _classMetadata;
      }
      // get all this assignments in the constructor
      var classMetadata = (0, _components.evaluateClassConstructor)(this.realm, componentType, props, context);
      this.realm.react.classComponentMetadata.set(componentType, classMetadata);
      return classMetadata;
    }
  }, {
    key: "_renderRelayQueryRendererComponent",
    value: function _renderRelayQueryRendererComponent(componentType, reactElement, props, context, evaluatedNode) {
      var renderResult = {
        result: reactElement,
        childContext: context
      };

      if (props instanceof _index.ObjectValue || props instanceof _index.AbstractObjectValue) {
        // get the "render" method off the instance
        var renderProp = (0, _index2.Get)(this.realm, props, "render");
        if (renderProp instanceof _index.ECMAScriptSourceFunctionValue && renderProp.$Call) {
          // if the render prop function is self contained, we can make it a new component tree root
          // and this also has a nice side-effect of hoisting the function up to the top scope
          if ((0, _utils.isRenderPropFunctionSelfContained)(this.realm, componentType, renderProp, this.logger)) {
            this._queueNewComponentTree(renderProp, evaluatedNode, true);
            return renderResult;
          } else {
            // we don't have nested additional function support right now
            // but the render prop is likely to have references to other components
            // that we need to also evaluate. given we can't find those components
            return renderResult;
          }
        }
      }
      // this is the worst case, we were unable to find the render prop function
      // and won't be able to find any further components to evaluate as trees
      // because of that
      return renderResult;
    }
  }, {
    key: "_renderComponent",
    value: function _renderComponent(componentType, props, context, branchStatus, branchState, evaluatedNode) {
      if ((0, _utils.valueIsKnownReactAbstraction)(this.realm, componentType)) {
        (0, _invariant2.default)(componentType instanceof _index.AbstractValue);
        this._queueNewComponentTree(componentType, evaluatedNode);
        evaluatedNode.status = "NEW_TREE";
        throw new _errors2.NewComponentTreeBranch();
      }
      (0, _invariant2.default)(componentType instanceof _index.ECMAScriptSourceFunctionValue);
      var value = void 0;
      var childContext = context;

      // first we check if it's a legacy class component
      if ((0, _utils.valueIsLegacyCreateClassComponent)(this.realm, componentType)) {
        throw new _errors2.ExpectedBailOut("components created with create-react-class are not supported");
      } else if ((0, _utils.valueIsClassComponent)(this.realm, componentType)) {
        var classMetadata = this._getClassComponentMetadata(componentType, props, context);
        var instanceProperties = classMetadata.instanceProperties,
            instanceSymbols = classMetadata.instanceSymbols;

        // if there were no this assignments we can try and render it as a simple class component

        if (instanceProperties.size === 0 && instanceSymbols.size === 0) {
          // We first need to know what type of class component we're dealing with.
          // A "simple" class component is defined as:
          //
          // - having only a "render" method
          // - having no lifecycle events
          // - having no state
          // - having no instance variables
          //
          // the only things a class component should be able to access on "this" are:
          // - this.props
          // - this.context
          // - this._someRenderMethodX() etc
          //
          // Otherwise, the class component is a "complex" one.
          // To begin with, we don't know what type of component it is, so we try and render it as if it were
          // a simple component using the above heuristics. If an error occurs during this process, we assume
          // that the class wasn't simple, then try again with the "complex" heuristics.
          try {
            value = this._renderSimpleClassComponent(componentType, props, context, branchStatus, branchState);
            this.simpleClassComponents.add(value);
          } catch (error) {
            // if we get back a SimpleClassBailOut error, we know that this class component
            // wasn't a simple one and is likely to be a complex class component instead
            if (error instanceof _errors2.SimpleClassBailOut) {
              // the component was not simple, so we continue with complex case
            } else {
              // else we rethrow the error
              throw error;
            }
          }
        }
        // handle the complex class component if there is not value
        if (value === undefined) {
          value = this._renderComplexClassComponent(componentType, props, context, classMetadata, branchStatus, branchState, evaluatedNode);
        }
      } else {
        value = this._renderFunctionalComponent(componentType, props, context);
        if ((0, _utils.valueIsFactoryClassComponent)(this.realm, value)) {
          (0, _invariant2.default)(value instanceof _index.ObjectValue);
          if (branchStatus !== "ROOT") {
            throw new _errors2.ExpectedBailOut("non-root factory class components are not suppoted");
          } else {
            // TODO support factory components
            return {
              result: value,
              childContext: childContext
            };
          }
        }
      }
      (0, _invariant2.default)(value !== undefined);
      return {
        result: this._resolveDeeply(componentType, value, context, branchStatus === "ROOT" ? "NO_BRANCH" : branchStatus, branchState, evaluatedNode),
        childContext: childContext
      };
    }
  }, {
    key: "_getRenderStrategy",
    value: function _getRenderStrategy(value) {
      // check if it's a ReactRelay.QueryRenderer
      if (this.realm.fbLibraries.reactRelay !== undefined) {
        var QueryRenderer = (0, _index2.Get)(this.realm, this.realm.fbLibraries.reactRelay, "QueryRenderer");
        if (value === QueryRenderer) {
          return "RELAY_QUERY_RENDERER";
        }
      } else if (value === (0, _utils.getReactSymbol)("react.fragment", this.realm)) {
        return "FRAGMENT";
      }
      return "NORMAL";
    }
  }, {
    key: "_resolveDeeply",
    value: function _resolveDeeply(componentType, value, context, branchStatus, branchState, evaluatedNode) {
      var _this2 = this;

      if (value instanceof _index.StringValue || value instanceof _index.NumberValue || value instanceof _index.BooleanValue || value instanceof _index.NullValue || value instanceof _index.UndefinedValue) {
        // terminal values
        return value;
      } else if (value instanceof _index.AbstractValue) {
        var length = value.args.length;
        if (length > 0) {
          var newBranchState = new _branching.BranchState();
          // TODO investigate what other kinds than "conditional" might be safe to deeply resolve
          for (var i = 0; i < length; i++) {
            value.args[i] = this._resolveDeeply(componentType, value.args[i], context, "NEW_BRANCH", newBranchState, evaluatedNode);
          }
          newBranchState.applyBranchedLogic(this.realm, this.reactSerializerState);
        }
        return value;
      }
      // TODO investigate what about other iterables type objects
      if (value instanceof _index.ArrayValue) {
        this._resolveArray(componentType, value, context, branchStatus, branchState, evaluatedNode);
        return value;
      }
      if (value instanceof _index.ObjectValue && (0, _utils.isReactElement)(value)) {
        // we call value reactElement, to make it clearer what we're dealing with in this block
        var reactElement = value;
        var typeValue = (0, _index2.Get)(this.realm, reactElement, "type");
        var propsValue = (0, _index2.Get)(this.realm, reactElement, "props");
        var refValue = (0, _index2.Get)(this.realm, reactElement, "ref");

        var resolveChildren = function resolveChildren() {
          // terminal host component. Start evaluating its children.
          if (propsValue instanceof _index.ObjectValue && propsValue.properties.has("children")) {
            var childrenValue = (0, _utils.getProperty)(_this2.realm, propsValue, "children");

            if (childrenValue instanceof _index.Value) {
              var resolvedChildren = _this2._resolveDeeply(componentType, childrenValue, context, branchStatus, branchState, evaluatedNode);
              // we can optimize further and flatten arrays on non-composite components
              if (resolvedChildren instanceof _index.ArrayValue) {
                resolvedChildren = (0, _utils.flattenChildren)(_this2.realm, resolvedChildren);
              }
              if (propsValue.properties.has("children")) {
                propsValue.refuseSerialization = true;
                _singletons.Properties.Set(_this2.realm, propsValue, "children", resolvedChildren, true);
                propsValue.refuseSerialization = false;
              }
            }
          }
          return reactElement;
        };

        if (typeValue instanceof _index.StringValue) {
          return resolveChildren();
        }
        // we do not support "ref" on <Component /> ReactElements
        if (!(refValue instanceof _index.NullValue)) {
          (0, _invariant2.default)(typeValue instanceof _index.ECMAScriptSourceFunctionValue || typeValue instanceof _index.AbstractObjectValue);
          var evaluatedChildNode = (0, _utils.createReactEvaluatedNode)("BAIL-OUT", (0, _utils.getComponentName)(this.realm, typeValue));
          evaluatedNode.children.push(evaluatedChildNode);
          this._queueNewComponentTree(typeValue, evaluatedChildNode);
          this._assignBailOutMessage(reactElement, "Bail-out: refs are not supported on <Components />");
          return reactElement;
        }
        if (!(propsValue instanceof _index.ObjectValue || propsValue instanceof _index.AbstractObjectValue || propsValue instanceof _index.AbstractValue)) {
          this._assignBailOutMessage(reactElement, "Bail-out: props on <Component /> was not not an ObjectValue or an AbstractValue");
          return reactElement;
        }
        var renderStrategy = this._getRenderStrategy(typeValue);

        if (renderStrategy === "NORMAL" && !(typeValue instanceof _index.ECMAScriptSourceFunctionValue || (0, _utils.valueIsKnownReactAbstraction)(this.realm, typeValue))) {
          this._assignBailOutMessage(reactElement, "Bail-out: type on <Component /> was not a ECMAScriptSourceFunctionValue");
          return reactElement;
        } else if (renderStrategy === "FRAGMENT") {
          return resolveChildren();
        }
        try {
          var result = void 0;
          switch (renderStrategy) {
            case "NORMAL":
              {
                (0, _invariant2.default)(typeValue instanceof _index.ECMAScriptSourceFunctionValue || typeValue instanceof _index.AbstractObjectValue);
                var _evaluatedChildNode = (0, _utils.createReactEvaluatedNode)("INLINED", (0, _utils.getComponentName)(this.realm, typeValue));
                evaluatedNode.children.push(_evaluatedChildNode);
                var render = this._renderComponent(typeValue, propsValue, context, branchStatus === "NEW_BRANCH" ? "BRANCH" : branchStatus, null, _evaluatedChildNode);
                result = render.result;
                this.statistics.inlinedComponents++;
                break;
              }
            case "RELAY_QUERY_RENDERER":
              {
                (0, _invariant2.default)(typeValue instanceof _index.AbstractObjectValue);
                var _evaluatedChildNode2 = (0, _utils.createReactEvaluatedNode)("RENDER_PROPS", (0, _utils.getComponentName)(this.realm, typeValue));
                evaluatedNode.children.push(_evaluatedChildNode2);
                var _render = this._renderRelayQueryRendererComponent(componentType, reactElement, propsValue, context, _evaluatedChildNode2);
                result = _render.result;
                break;
              }
            default:
              (0, _invariant2.default)(false, "unsupported render strategy");
          }

          if (result instanceof _index.UndefinedValue) {
            this._assignBailOutMessage(reactElement, "Bail-out: undefined was returned from render");
            if (branchStatus === "NEW_BRANCH" && branchState) {
              return branchState.captureBranchedValue(typeValue, reactElement);
            }
            return reactElement;
          }
          if (branchStatus === "NEW_BRANCH" && branchState) {
            return branchState.captureBranchedValue(typeValue, result);
          }
          return result;
        } catch (error) {
          // assign a bail out message
          if (error instanceof _errors2.NewComponentTreeBranch) {
            // NO-OP (we don't queue a newComponentTree as this was already done)
          } else {
            (0, _invariant2.default)(typeValue instanceof _index.ECMAScriptSourceFunctionValue || typeValue instanceof _index.AbstractObjectValue);
            var _evaluatedChildNode3 = (0, _utils.createReactEvaluatedNode)("BAIL-OUT", (0, _utils.getComponentName)(this.realm, typeValue));
            evaluatedNode.children.push(_evaluatedChildNode3);
            this._queueNewComponentTree(typeValue, _evaluatedChildNode3);
            if (error instanceof _errors2.ExpectedBailOut) {
              this._assignBailOutMessage(reactElement, "Bail-out: " + error.message);
            } else if (error instanceof _errors.FatalError) {
              this._assignBailOutMessage(reactElement, "Evaluation bail-out");
            } else {
              throw error;
            }
          }
          // a child component bailed out during component folding, so return the function value and continue
          if (branchStatus === "NEW_BRANCH" && branchState) {
            return branchState.captureBranchedValue(typeValue, reactElement);
          }
          return reactElement;
        }
      } else {
        throw new _errors2.ExpectedBailOut("unsupported value type during reconcilation");
      }
    }
  }, {
    key: "_assignBailOutMessage",
    value: function _assignBailOutMessage(reactElement, message) {
      // $BailOutReason is a field on ObjectValue that allows us to specify a message
      // that gets serialized as a comment node during the ReactElement serialization stage
      if (reactElement.$BailOutReason !== undefined) {
        // merge bail out messages if one already exists
        reactElement.$BailOutReason += ", " + message;
      } else {
        reactElement.$BailOutReason = message;
      }
    }
  }, {
    key: "_resolveArray",
    value: function _resolveArray(componentType, arrayValue, context, branchStatus, branchState, evaluatedNode) {
      var _this3 = this;

      (0, _utils.forEachArrayValue)(this.realm, arrayValue, function (elementValue, elementPropertyDescriptor) {
        elementPropertyDescriptor.value = _this3._resolveDeeply(componentType, elementValue, context, branchStatus, branchState, evaluatedNode);
      });
    }
  }]);

  return Reconciler;
}();
//# sourceMappingURL=reconcilation.js.map